from .veritas_analysis import VeritasAnalysis
import logging
try:
	from .fermi_analysis import FermiAnalysis
	from .joint_analysis import JointAnalysis
except:

	logging.warning("Fermitools is not installed. Any Fermi-LAT related analysis cannot be performed.")


import numpy as np
import matplotlib.pyplot as plt

import astropy.units as u

from .version import __version__

from . import utils
from .config import JointConfig
from .download import DownloadFermiData
from .analysis import *

from astropy.visualization import astropy_mpl_style, quantity_support

plt.style.use(astropy_mpl_style)

quantity_support()



import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
import os
import pickle
import copy
import glob

from . import FermiAnalysis, VeritasAnalysis
from ..utils import logger
from .. import utils
from ..model import default_model
from .. import plotting
from ..config import JointConfig

from gammapy.datasets import Datasets, FluxPointsDataset
from gammapy.modeling import Fit
import gammapy.modeling.models as gammapy_model
from gammapy.estimators import FluxPoints, FluxPointsEstimator
from gammapy.modeling.models import SkyModel


class JointAnalysis:
    """
    This is to perform a joint VERITAS and Fermi-LAT analysis. This class
    exploits results from VeritasAnalysis and FermiAnalysis.

    Args:
        veritas (str or vtspy.VeritasAnalysis): state filename or class for VERITAS
            Default: initial
        fermi (str or vtspy.FermiAnalysis): state filename or class for Fermi-LAT
            Default: initial
        config_file (str): config filename (yaml)
            Default: config.yaml
        verbosity (int)
    """


    def __init__(self, veritas = None, fermi = None, config_file='config.yaml', verbosity=1, **kwargs):
        self._verbosity = verbosity
        self._logging = logger(self.verbosity)
        self._logging.info("Initialize the joint-fit analysis...")
        self.config = JointConfig.get_config(config_file).pop("joint")
        self._outdir = kwargs.pop("outdir", self.config["fileio"]["outdir"])

        if not(os.path.isdir(self._outdir)):
            os.system(f"mkdir {self._outdir}")

        self._model_change_flag = False
        self._fit_flag = False

        if type(veritas) == str:
            self.veritas = VeritasAnalysis(veritas, config_file=config_file)
            self._veritas_state = veritas
            self._target_name = self.veritas.target_name
        elif hasattr(veritas, "datasets"):
            self._logging.info("VERITAS datasets is imported.")
            self.veritas = veritas
            self._target_name = self.veritas.target_name

        if type(fermi) == str:
            self.fermi = FermiAnalysis(fermi, construct_dataset=True, config_file=config_file)
            self._fermi_state = fermi
        elif hasattr(fermi, "gta"):
            self.fermi = fermi
            self.fermi.construct_dataset()
            self._logging.info("Fermi-LAT datasets is imported.")

        if hasattr(self, "fermi") and hasattr(self, "veritas"):
            self._logging.info("Constructing a joint datasets")
            self._construct_joint_datasets()
        self._logging.info("Completed.")

    @property
    def target_model(self):
        """
        Return:
            gammapy.modeling.models.SkyModel
        """
        return self.datasets.models[self.target_name]

    @property
    def verbosity(self):
        """
        Return:
            int
        """
        return self._verbosity

    @property
    def target_name(self):
        """
        Return:
            str: target name
        """
        return self._target_name

    def print_datasets(self):
        """
        Print datasets

        Return:
            astropy.table
        """

        return self._logging.info(self.datasets)


    def print_models(self, full_output=False):
        """
        Print model and parameters

        Args:
            full_output (bool): return a target model or all models
                default: False
        Return:
            astropy.table
        """
        if full_output:
            return self._logging.info(self.datasets.models)
        else:
            table = self.datasets.models.to_parameters_table()
            table = table[table["model"]== self.target_name]
            return table

    def save_state(self, state_file):
        """
        Save the state

        Args:
            state_file (str): the name of state
        """

        filename = f"./{self._outdir}/{state_file}.pickle".format(state_file)
        with open(filename, 'wb') as file:
            temp = [copy.copy(self.veritas), copy.copy(self.fermi)]
            del(self.veritas)
            del(self.fermi)
            del(self._logging)
            pickle.dump(self, file)
            self._logging = logger(self.verbosity)
            self.veritas, self.fermi = temp

    def load_state(self, state_file):
        """
        Load the state

        Args:
        state_file (str): the name of state
        """
        try:
            filename = f"./{self._outdir}/{state_file}.pickle".format(state_file)
            with open(filename, 'rb') as file:
                self.__dict__.update(pickle.load(file).__dict__)

            if not(hasattr(self, "fermi")):
                self.fermi = FermiAnalysis(self._fermi_state, construct_dataset=True)

            if not(hasattr(self, "veritas")):
                self.veritas = VeritasAnalysis(self._veritas_state)

            self._construct_joint_datasets()


        except:
            self._logging.error("The state file does not exist. Check the name again")
            return -1

    def list_of_state(self):
        files = glob.glob(f"{self._outdir}/*.pickle")
        return print([n.split("/")[-1].split(".")[0] for n in files])

    def optimize(self, method="flux", model=None, instrument="VERITAS", **kwargs):
        """
        To find good initial parameters for a given model.

        Args:
            method (str): either "flux" or "inst". The "flux" method will fit flux points
            from VeritasAnalysis.analysis("sed") and FermiAnalysis.analysis("sed"). The
            "inst" method will fit the model with one of datasets (defined by instrument)
            model (gammapy.modeling.models.SpectralModel): a model to fit
                Default: None (fit with the current target model)
            instrument(str): instrument used for the "inst" method
                Default: VERITAS
            **kwargs
        """

        if model is None:
            model = self.datasets.models[self.target_name].spectral_model

        if method == "flux":
            test_model = SkyModel(spectral_model=model, name="test")
            fermi_sed = kwargs.pop("fermi_sed", f"{self.fermi._outdir}/{self._fermi_state}_sed.fits")

            if not(os.path.isfile(fermi_sed)):
                self.fermi.analysis("sed", state_file = self._fermi_state)

            data = FluxPoints.read(fermi_sed, reference_model=self.target_model, hdu=1)

            fermi_dataset = FluxPointsDataset(data=data, models=test_model)
            fermi_dataset.mask_safe = ~fermi_dataset.data.to_table()["is_ul"]

            veritas_dataset = FluxPointsDataset(data=self.veritas.flux_points, models=test_model)
            nan_norm = ~np.isnan(veritas_dataset.data.to_table()["norm"])
            veritas_dataset.mask_safe = veritas_dataset.mask_safe*nan_norm

            datasets = Datasets([fermi_dataset, veritas_dataset])
            datasets.models = test_model
            
            optimize_opts_default = {
                "method": "L-BFGS-B",
                "options": {"ftol": 1e-4, "gtol": 1e-05, "maxls": 40},
                "backend": "scipy",
            }
            optimize_opts = kwargs.pop("optimize_opts", optimize_opts_default)

            fit_ = Fit(optimize_opts =  optimize_opts)
            result_optimize = fit_.run(datasets)
            self._logging.info(result_optimize)
            self.datasets.models[self.target_name].spectral_model = test_model.spectral_model
        elif method == "inst":
            self.datasets.models[self.target_name].spectral_model = model
            joint_fit = Fit()
            fit_results = joint_fit.run(self.datasets[instrument.lower()])
        else:
            return

    def fit(self, **kwargs):
        """
        Perform a joint-fit analysis

        Args:
            **kwargs: passed to vtspy.JointAnalysis.optimize
        """

        optimize = kwargs.pop("optimize", True)
        if self._model_change_flag and optimize:
            self._logging.info("A model is recently updated. Optimizing the input parameters...")
            self.optimize(**kwargs)
            self._model_change_flag = False
            self._logging.info("Completed. Move to the next step.")

        self._logging.info("Start fitting...")

        joint_fit = Fit()
        self.fit_results = joint_fit.run(self.datasets)

        if self.fit_results.success:
            self._logging.info("Fit successfully.")
            self._fit_flag = True
        else:
            self._logging.error("Fit fails.")

    def analysis(self, **kwargs):

        """
        Perform a SED analysis.

        Args:
            **kwargs: passed to FluxPointsEstimator
        """

        energy_bins = kwargs.get("energy_bins", np.geomspace(0.0001, 10, 20) * u.TeV)

        fpe = FluxPointsEstimator(
            energy_edges=energy_bins,
            source=self.target_name, selection_optional="all"
            )

        self.flux_points = fpe.run(self.datasets)

        self._flux_points_dataset = FluxPointsDataset(
            data=self.flux_points, models=self.datasets.models
        )


    def plot(self, output, **kwargs):
        """
        Show various results: SED

        Args:
            output (str): a plot to show
                Options: ["sed"]
            **kwargs: passed to vtspy.JointAnalysis.plot_sed
        """

        if output == "sed":
            self.plot_sed(**kwargs)

    def plot_sed(self, fermi=True, veritas=True, joint=True, show_model = True, show_flux_points=True):
        """
        Plot a spectral energy distribution (SED) with a model and flux points.

        Args:
            fermi (bool): show Fermi-LAT results
                Default: True
            veritas (bool): show VERITAS results
                Default: True
            fermi (bool): show Joint-fit results
                Default: True
            show_flux_points (bool): slow flux points
                Default: True
            show_model (bool) show model
                Default: True
        """

        if fermi and not(hasattr(self.fermi, "output")):
            self.fermi.simple_analysis("sed")

        if veritas and not(hasattr(self.veritas, "_flux_points_dataset")):
            self.veritas.simple_analysis()

        if joint and not(self._fit_flag):
            fit = False
        else:
            fit = True

        cmap = plt.get_cmap("tab10")
        i = 0

        if joint and show_model:
            if fit:
                energy_bounds = [100 * u.MeV, 30 * u.TeV]
                jf_model = self.datasets.models[0].spectral_model
                jf_model.plot(energy_bounds=energy_bounds, sed_type="e2dnde", color=cmap(i), label=self.target_name)
                jf_model.plot_error(energy_bounds=energy_bounds,
                                         sed_type="e2dnde", alpha=0.2, color="k")
            else:
                energy_bounds = [100 * u.MeV, 30 * u.TeV]
                jf_model = self.datasets.models[self.target_name].spectral_model

                if fit:
                    jf_model.plot(energy_bounds=energy_bounds, sed_type="e2dnde", color=cmap(i), label=self.target_name, ls="-")
                    jf_model.plot_error(energy_bounds=energy_bounds,
                                         sed_type="e2dnde", alpha=0.2, color="k")
                else:
                    jf_model.plot(energy_bounds=energy_bounds, sed_type="e2dnde", color=cmap(i), label="Before fit", ls="--")
            i+=1

        if veritas:
            vts = self.veritas._flux_points_dataset
            energy_bounds = vts._energy_bounds
            if show_flux_points:
                vts.data.plot(sed_type="e2dnde", color = cmap(i), label="VERITAS")

            if not(fit) and show_model:
                veritas_model = vts.models[0].spectral_model
                veritas_model.plot(energy_bounds=energy_bounds, sed_type="e2dnde", color=cmap(i))
                veritas_model.plot_error(energy_bounds=energy_bounds,
                                         sed_type="e2dnde", alpha=0.2, facecolor=cmap(i))
            i+=1

        if fermi:
            plotting.plot_sed(self.fermi.output, units="TeV", erg=True, show_flux_points=show_flux_points,
                show_model = not(fit)*show_model, color=cmap(i))

            plt.xlim(5e-5, 30)
            i+=1

        plt.xscale("log")
        plt.yscale("log")
        plt.legend(fontsize=13)
        plt.grid(which="major", ls="-")
        plt.grid(which="minor", ls=":")
        plt.xlabel("Energy [TeV]", fontsize=13)
        plt.ylabel(r"Energy flux [erg/cm$^2$/s]", fontsize=13)


    def change_model(self, model, optimize=False, **kwargs):
        """
        To change a target model

        Args:
            model (str or gammapy.modeling.models.SpectralModel): a new target model
            optimize (bool): perform optimization (JointAnalysis.optimize)
                Default: False
            **kwargs: passed to JointAnalysis.optimize
        """

        prevmodel = self.datasets.models[self.target_name].spectral_model.tag[0]
        if type(model) == str:
            spectral_model = default_model(model, **kwargs)
            if model is None:
                self._logging.error("The input model is not supported yet.")
                return
        elif hasattr(model, "tag"):
            spectral_model = model
            self._logging.info(f"A model, {model.tag[0]}, is imported")

        if optimize:
            self.optimize(model=spectral_model, **kwargs)
            self._model_change_flag = False
        else:
            self._model_change_flag = True
            self.datasets.models[self.target_name].spectral_model = spectral_model

        newmodel = self.datasets.models[self.target_name].spectral_model.tag[0]
        self._fit_flag = False
        self._logging.info(f"The spectral model for the target is chaged:")
        self._logging.info(f"{prevmodel}->{newmodel}")

    def add_dataset(self, data, sync = True, **kwargs):
        if type(data) == FluxPoints:
            if sync:
                model = self.target_model
            else:
                self._logging.warning(f"The model is assumed to be a power law.")
                model = SkyModel(spectral_model=gammapy_model.PowerLawSpectralModel(), name="test")
            new_dataset = FluxPointsDataset(data=data, models=model, **kwargs)
            self.datasets.append(new_dataset)
        else:
            self._logging.error(f"This data type is not supported yet.")
            
    def _construct_joint_datasets(self, inst="VERITAS"):
        vts_model = self.veritas.stacked_dataset.models[0]
        fermi_model = self.fermi.datasets.models[self.fermi.target_name]

        self.veritas.stacked_dataset.models = self._find_target_model()
        self.datasets = Datasets([self.fermi.datasets, self.veritas.stacked_dataset])

        if inst.lower() == "veritas":
            self.datasets.models[self.fermi.target_name].spectral_model = vts_model.spectral_model
        elif inst.lower() == "fermi":
            self.datasets.models[self.veritas.target_name].spectral_model = fermi_model.spectral_model

        self.datasets.models[self.fermi.target_name]._name = self.target_name


    def _find_target_model(self):
        target_pos = self.fermi.datasets.models[self.fermi.target_name].spatial_model.position
        th2cut = self.veritas._on_region.radius.value

        models = []
        for model in self.fermi.datasets.models:
            if model.name != 'galdiff' and model.name != 'isodiff':
                if target_pos.separation(model.spatial_model.position).deg < th2cut:
                    models.append(model)
        return models



import numpy as np
import os, sys
import urllib
import time
import yaml
import html2text
from pathlib import Path

from astropy.io import fits

import re

from .config import JointConfig
from .utils import logger

class DownloadFermiData:
    
    def __init__(self, config_file = "config.yaml", dtype="Photon", verbosity = 1):
        """
        This is to download Fermi-LAT data based on the configuration
        file generated by the 'FermipyConfig' class. 

        Args:
            config_file (str): Fermi config filename (yaml)
                Default: config.yaml 
            dtype (str): either Photon or Extended
                Default: Photon
            verbosity (int)
        """
        self._logging = logger(verbosity=verbosity)

        self.config_file = config_file
        self.config = JointConfig.get_config(self.config_file).pop("fermi")
        path = Path(self.config["data"]["evfile"])
        self.datadir = str(path.parent)
        self.outdir = self.config["fileio"]["outdir"]

        if self.config['selection']['target'] == None:
            self.target = "source"
        else:
            self.target = self.config['selection']['target']

        self.coordsys = self.config['binning']['coordsys']
        if self.coordsys == "GAL":
            self.coord = "Galactic"
            self.loc = [self.config['selection']['glon'], self.config['selection']['glat']]
        elif self.coordsys == "CEL":
            self.coord = "J2000"
            self.loc = [self.config['selection']['ra'], self.config['selection']['dec']]
        
        if self.loc[0] == None or self.loc[1] == None:
            self._logging.error("[Error] Coordinates (e.g., RA & DEC) is not specified.")
            return
        
        self.tmin = self.config['selection']['tmin']
        self.tmax = self.config['selection']['tmax']    
        if self.tmin == None or self.tmax == None:
            self._logging.error("[Error] Time range is not specfied.")
            return

        self.emin = min(100, self.config['selection']['emin'])
        self.emax = max(300000, self.config['selection']['emax'])
        self.dtype = dtype

        success = self._query()

        if success:
            self._download(datadir=self.datadir)


    def _query(self):

        url                         = "https://fermi.gsfc.nasa.gov/cgi-bin/ssc/LAT/LATDataQuery.cgi"
        parameters                  = {}
        parameters['coordfield']    = "%s,%s" %(self.loc[0], self.loc[1])
        parameters['coordsystem']   = "%s" %(self.coord)
        parameters['shapefield']    = "%s" %(15)
        parameters['timefield']     = "%s,%s" %(self.tmin,self.tmax)
        parameters['timetype']      = "%s" %("MET")
        parameters['energyfield']   = "%s,%s" %(self.emin,self.emax)
        parameters['photonOrExtendedOrNone'] = "%s" %(self.dtype)
        parameters['destination']   = 'query'
        parameters['spacecraft']    = 'checked'

        self._logging.info("Query parameters:")
        for k,v in parameters.items():
            self._logging.info("%30s = %s" %(k,v))

        postData                    = urllib.parse.urlencode(parameters).encode("utf-8")
        temporaryFileName           = "__temp_query_result.html"
        try:
            os.remove(temporaryFileName)
        except:
            pass
        pass

        urllib.request.urlcleanup()

        urllib.request.urlretrieve(url, temporaryFileName, lambda x,y,z:0, postData)

        with open(temporaryFileName) as htmlFile:
            lines = []
            for line in htmlFile:
                lines.append(line.encode('utf-8'))

            html = "".join(str(lines)).strip()
        
        self._logging.debug("Answer from the LAT data server:")
        
        text = html2text.html2text(html.strip()).split("\n")
        text = list(filter(lambda x:x.find("[") < 0 and  x.find("]") < 0 and x.find("#") < 0 and x.find("* ") < 0 and
                        x.find("+") < 0 and x.find("Skip navigation")<0,text))
        text = list(filter(lambda x:len(x.replace(" ",""))>1,text))
        text = [t for t in text if t[0] != '\\']

        
        for t in text:
            if "occurs after data end MET" in t:
                maxTime = re.findall("occurs after data end MET \(([0-9]+)\)", t)[0]
                self._logging.error("[Error] The current Fermi Data Server does not have data upto the entered 'tmax'.")
                self._logging.error("[Error] 'tmax' value in the config file is changed to the maximum value.")
                self._logging.error("[Error] config['selection']['tmax'] = ", maxTime)
                self._logging.error("[Error] Please try again.")
                self.config['selection']['tmax'] = float(maxTime)
                JointConfig.updateConfig(self.config, self.config_file)
                return False


        text[-3] = text[-3]+" "+text[-2]
        text.remove(text[-2])
        text[-2] = text[-2]+text[-1]
        text.remove(text[-1])

        for t in text: self._logging.debug(t)

        os.remove(temporaryFileName)
        for ln, t in enumerate(text):
            estimatedTimeForTheQuery = re.findall("The estimated time for your query to complete is ([0-9]+) seconds",t)
            if len(estimatedTimeForTheQuery)>0:
                estimatedTimeForTheQuery = estimatedTimeForTheQuery[0]
                break

        for ln, t in enumerate(text):
            address = re.findall("your query may be found ([a-z]+) ",t)
            if len(address)>0:
                line_number = ln
                break

        address = text[line_number] + text[line_number+1]
        self.httpAddress = address.split()[-2][1:-2]
        
        startTime = time.time()
        timeout = 2.*max(5.0,float(estimatedTimeForTheQuery))
        regexpr = re.compile("wget (.*.fits)")

        links = None
        fakeName = "__temp__query__result.html"

        overTime = False

        self._logging.info("The estimated time is about "+str(int(estimatedTimeForTheQuery))+" seconds.")
        while(time.time() <= startTime+timeout):
            remainedTime = int(int(estimatedTimeForTheQuery) - (time.time()-startTime))
            if remainedTime<0 and not(overTime):
                overTime=True
                self._logging.info("The Fermi data is still not ready. Wait for another " + str(int(estimatedTimeForTheQuery)) + " seconds.")
            try:
                (filename, header) = urllib.request.urlretrieve(self.httpAddress,fakeName)
            except:
                urllib.request.urlcleanup()
                continue
            
            with open(fakeName) as f:
                html = " ".join(f.readlines())
                try:
                    status = re.findall("The state of your query is ([0-9]+)",html)[0]
                except:
                    status = '0'
                    pass

                if(status=='2'):
                    links = regexpr.findall(html)
                    if len(links) >= 2:
                        break
                
            os.remove(fakeName)
            urllib.request.urlcleanup()

        try:
            os.remove(fakeName)
        except:
            self._logging.error("[Error] The files (SC and EV files) are not ready to be downloaded. Check the link and then use 'DownloadFermiData.manualDownload()' when the data is ready.")
            return
        
        np.save(f"{self.outdir}/fermi_dwn_link", links)

        return True

    def _download(self, datadir="./fermi/"):

        links = np.load(f"{self.outdir}/fermi_dwn_link.npy")

        for lk in links:
            self._logging.info("Downloading... "+lk)
            fileName = lk[-9:-5]

            if "SC" in lk:
                self.config['data']['scfile'] = f"{datadir}/{fileName}.fits"
            else:
                with open(f"{datadir}/EV00.lst", "a") as f:
                    f.write(fileName+".fits\n")

            urllib.request.urlretrieve(lk, f"{datadir}/{fileName}.fits")

        self._logging.info("Downloading the Fermi-LAT data has been completed.")
        os.system(f"rm {self.outdir}/fermi_dwn_link.npy")
        JointConfig.update_config(self.config, "fermi", self.config_file)

    def manual_download(self, address=None):
        """
        When an error occurs, one can manually download the data
        
        Args:
            address (str, optional): address to the Fermi-LAT data page.
        """
        if address!=None:
            self.httpAddress = address

        links = None
        fakeName = "__temp__query__result.html"
        regexpr = re.compile("wget (.*.fits)")

        (filename, header) = urllib.request.urlretrieve(self.httpAddress,fakeName)
        
        with open(fakeName) as f:
            html = " ".join(f.readlines())
            links = regexpr.findall(html)
        
        os.remove(fakeName)
        urllib.request.urlcleanup()
        
        np.save(f"{self.outdir}/fermi_dwn_link", links)

        self._download(datadir=self.datadir)





import numpy as np
import astropy.units as u
from astropy.constants import c
from astropy.coordinates import Distance

from agnpy.spectra import BrokenPowerLaw
from agnpy.synchrotron import Synchrotron
from agnpy.compton import SynchrotronSelfCompton

from gammapy.modeling import Parameter
from gammapy.modeling.models import (
    SpectralModel,
    SPECTRAL_MODEL_REGISTRY,
)


class agnpy_spectral_model(SpectralModel):
    """
    Wrapper of agnpy's synchrotron and SSC classes. A broken power law is assumed
    for the electron spectrum. To limit the span of the parameters space, we fit
    the log10 of the parameters whose range is expected to cover several orders
    of magnitudes (normalisation, gammas, size and magnetic field of the blob).
    """

    tag = ["agnpy(SYN+SSC)", "agnpy"]

    # electron spectrum parameters
    log10_norm_e = Parameter("log10_norm_e", "-5", min=-20, max=10)
    p1 = Parameter("p1", 2.1, min=-2.0, max=5.0)
    p2 = Parameter("p2", 3.1, min=-2.0, max=5.0)

    # Lorentz factor
    log10_gamma_b = Parameter("log10_gamma_b", 3, min=1, max=7)
    log10_gamma_min = Parameter("log10_gamma_min", 1, min=0, max=4)
    log10_gamma_max = Parameter("log10_gamma_max", 5, min=4, max=8)

    # emission region parameters
    delta_D = Parameter("delta_D", 10, min=0, max=60)
    log10_B = Parameter("log10_B", -1, min=-4, max=2)
    t_var = Parameter("t_var", "600 s", min=10, max=np.pi * 1e7)

    # source general parameters
    z = Parameter("z", 0.1, min=0.01, max=1)
    d_L = Parameter("d_L", "1e27 cm", min=1e25, max=1e33)

    @staticmethod
    def evaluate(
        energy,
        log10_norm_e,
        p1,
        p2,
        log10_gamma_b,
        log10_gamma_min,
        log10_gamma_max,
        delta_D,
        log10_B,
        t_var,
        z,
        d_L,
    ):
        # conversions
        norm_e = 10 ** log10_norm_e * u.Unit("cm-3")
        gamma_b = 10 ** log10_gamma_b
        gamma_min = 10 ** log10_gamma_min
        gamma_max = 10 ** log10_gamma_max
        B = 10 ** log10_B * u.G
        R_b = (c * t_var * delta_D / (1 + z)).to("cm")
        pars = (z, d_L, delta_D, B, R_b, BrokenPowerLaw, norm_e,
                p1, p2, gamma_b, gamma_min, gamma_max)
        pars = (z, d_L, delta_D, B, R_b, BrokenPowerLaw, norm_e,
                p1, p2, gamma_b, gamma_min, gamma_max)
        nu = energy.to("Hz", equivalencies=u.spectral())
        dim = len(energy.shape)

        if dim == 2:
            sed_synch = np.asarray([Synchrotron.evaluate_sed_flux(n,*pars) for n in nu.T])
            sed_ssc = np.asarray([SynchrotronSelfCompton.evaluate_sed_flux(n,*pars) for n in nu.T])
            sed_units = u.erg / u.cm / u.cm / u.second
            sed = (sed_synch.T + sed_ssc.T) * sed_units
        else:
            sed_synch = Synchrotron.evaluate_sed_flux(nu,*pars)
            sed_ssc = SynchrotronSelfCompton.evaluate_sed_flux(nu,*pars)
            sed = (sed_synch + sed_ssc)
        return (sed / energy ** 2).to("1 / (cm2 TeV s)")



import os
import numpy as np
import matplotlib.pyplot as plt
import glob
import pickle
import copy
import re
import astropy.units as u
from astropy.coordinates import SkyCoord, Angle
from astropy.time import Time
from astropy.io import fits

from ..utils import logger, LiMaSiginficance
from .. import utils
from ..model import default_model
from ..plotting import plot_ROI
from ..config import JointConfig

from gammapy.data import DataStore
from gammapy.datasets import Datasets, SpectrumDataset, FluxPointsDataset

from gammapy.maps import WcsGeom, MapAxis, RegionGeom

from gammapy.makers import (
    SafeMaskMaker,
    SpectrumDatasetMaker,
    ReflectedRegionsBackgroundMaker,
)

from gammapy.modeling import Fit

import gammapy.modeling.models as gammapy_model

from gammapy.modeling.models import SkyModel

from regions import CircleSkyRegion

from gammapy.estimators import FluxPointsEstimator, ExcessMapEstimator, FluxPoints, LightCurveEstimator



class VeritasAnalysis:

	def __init__(self, state_file = "initial", config_file='config.yaml', overwrite=False, verbosity=1, **kwargs):
		"""
	    This is to perform a simple VERITAS analysis and to prepare the joint-fit
	    analysis. All analysis is done with the gammapy package

	    Args:
	        state_file (str): state filename (pickle)
	        	Default: initial
	        config_file (str): config filename (yaml)
	            Default: config.yaml
	        overwrite (bool): overwrite the state
	            Default: False
	        verbosity (int)
	        **kwargs: passed to VeritasAnalysis.setup
	    """
		self._verbosity = verbosity

		self.config = JointConfig.get_config(config_file).pop("veritas")

		self._logging = logger(self.verbosity)

		self._logging.info("Initialize the VERITAS analysis.")

		self._datadir = self.config['data']['anasum']
		self._outdir = self.config['fileio']['outdir']

		self._eff_cut = self.config['cuts']['eff_cut']
		self._bias_cut = self.config['cuts']['bias_cut']
		self._max_region_number = self.config['selection']['max_region_number']

		self._energy_bins = MapAxis.from_bounds(self.config["selection"]["emin"], self.config["selection"]["emax"], nbin=self.config["selection"]["nbin"], interp="log", unit="TeV").edges
		self._energy_axis = MapAxis.from_energy_bounds(
		    1e-2, 1e4, nbin=10, per_decade=True, unit="TeV", name="energy"
		)
		self._energy_axis_true = MapAxis.from_energy_bounds(
		    1e-2, 1e4, nbin=5, per_decade=True, unit="TeV", name="energy_true"
		)
		self._excluded_regions = []

		if overwrite or not(os.path.isfile("./{}/{}.pickle".format(self._outdir, state_file))):

			self.setup(**kwargs)

			self._logging.debug("Save the initial state.")

			self.save_state(state_file, init=True)

			self._logging.info("The initial setup is saved [state_file = {}].".format(state_file))

		else:
			self._logging.info("The setup is found [state_file = {}]. Read the state.".format(state_file))
			flag = self.load_state(state_file)

			if flag == -1:
				return

		self._logging.info("Completed (VERITAS initialization).")

	@property
	def target(self):
		"""
        Return:
            astropy.SkyCoord
        """
		return SkyCoord(ra = self.config["selection"]["ra"],
						dec = self.config["selection"]["dec"],
                  		unit="deg", frame="icrs")

	@property
	def target_name(self):
		"""
        Return:
            str: target name
        """
		return self.config["selection"]["target"]

	@property
	def obs_ids(self):
		"""
        Return:
            list: list of observation id
        """
		return self._obs_ids

	@property
	def energy_bins(self):
		"""
        Return:
        	MapAxis.edges: energy bins used in flux points.
        """
		return self._energy_bins

	@property
	def verbosity(self):
		"""
        Return:
        	int
        """
		return self._verbosity

	def print_flux(self):
		"""
        Return:
        	astropy.table: flux points in SED
        """
		if hasattr(self, "flux_points"):
			return self.flux_points.to_table(sed_type="e2dnde", formatted=True)

	def print_lightcurve(self, sed_type='eflux'):
		"""
        Return:
        	astropy.table: flux points in lightcurve
        """
		if hasattr(self, "_lightcurve"):
			return self._lightcurve.to_table(sed_type='eflux',format='lightcurve')

	def print_models(self):
		"""
        Return:
        	astropy.table: fit parameters
        """
		if hasattr(self, "stacked_dataset"):
			return self.stacked_dataset.models.to_parameters_table()

	def peek_dataset(self):
		"""
        Show dataset information
        """
		self.stacked_dataset.peek()

	def save_state(self, state_file, init=False):
		"""
		Save the state

		Args:
		    state_file (str): the name of state
		    init (bool): check whether this is the initial analysis.
		        Default: False
		"""

		if (init==False) and (state_file == "initial"):
		    self._logging.warning("The 'inital' state is overwritten. This is not recommended.")
		    self._logging.warning("The original 'inital' state is archived in the '_orig' folder.")
		    os.system(f"mkdir ./{self._outdir}/_orig")
		    for file in glob.glob(f"./{self._outdir}/*initial*"):
		        os.sytem(f"mv {file} ./{self._outdir}/_orig/")

		filename = f"./{self._outdir}/{state_file}.pickle".format(state_file)
		with open(filename, 'wb') as file:
			del(self._logging)
			pickle.dump(self, file)
			self._logging = logger(self.verbosity)
			self._veritas_state = state_file

	def load_state(self, state_file):
		"""
		Load the state

		Args:
		state_file (str): the name of state
		"""
		try:
			filename = f"./{self._outdir}/{state_file}.pickle".format(state_file)
			with open(filename, 'rb') as file:
				self.__dict__.update(pickle.load(file).__dict__)
			self._veritas_state = state_file
		except:
			self._logging.error("The state file does not exist. Check the name again")
			return -1

	def setup(self, **kwargs):
		"""
	    This is to initialize the VERITAS analysis; e.g., construct datasets
	    To change the setting for this setup, check config file.

	    VeritasAnalysis.config

	    Args:
	        **kwargs: passed to VeritasAnalysis.construct_dataset
	    """

		selection = dict(
		    type="sky_circle",
		    frame="icrs",
		    lon= self.target.ra,
		    lat= self.target.dec,
		    radius = self.config["selection"]['radius'] * u.deg,
		)

		self._logging.info("Load the data files.")

		

		if not(os.path.exists(f"{self._datadir}/hdu-index.fits.gz")):
			self._logging.warning("The 'hdu-/obs-index' files is not found.")
			try:
				from pyV2DL3 import generateObsHduIndex
			except:
				self._logging.error("The pyV2DL3 package is required to proceed.")
				return

			import glob
			filelist = glob.glob(f"{self._datadir}/*anasum.fit*")
			generateObsHduIndex.create_obs_hdu_index_file(filelist, index_file_dir=self._datadir)
			self._logging.info("The hdu-index and obs-index files are created.")	
		else:
			rflag = self._quick_check_runlist()
			if not(rflag):
				self._logging.warning("A mismatch in the obs-index file is found.")
				try:
					from pyV2DL3 import generateObsHduIndex
				except:
					self._logging.error("The pyV2DL3 package is required to proceed.")
					return

				import glob
				filelist = glob.glob(f"{self._datadir}/*anasum.fit*")
				generateObsHduIndex.create_obs_hdu_index_file(filelist, index_file_dir=self._datadir)
				self._logging.info("The hdu-index and obs-index files are created.")

		self._data_store = DataStore.from_dir(f"{self._datadir}")

		self._obs_ids = self._data_store.obs_table.select_observations(selection)["OBS_ID"]

		self.observations = self._data_store.get_observations(self._obs_ids, required_irf=["aeff", "edisp"])
		time_intervals = [self.config["selection"]["tmin"], self.config["selection"]["tmax"]]
		self.observations, self._obs_ids = utils.time_filter(self.observations, time_intervals, time_format="mjd")
		self._logging.info(f"The number of observations is {len(self.observations)}")

		self._th2cut = self.config["cuts"]['th2cut']
		self._on_region = CircleSkyRegion(center=self.target, radius=Angle(np.sqrt(self._th2cut)*u.deg))

		self._logging.info("Define exclusion regions.")
		if kwargs.pop("simbad", self.config["background"]["simbad"]):
			self._exclusion_mask = self._exclusion_from_simbad(**kwargs)
		else:
			self._exclusion_mask = self._exclusion_from_bright_src_list(**kwargs)
		self.add_exclusion_region(coord=[self.target.ra, self.target.dec], radius=self.config["selection"]["exc_on_region_radius"])

		self._logging.info("Define ON- and OFF-regions.")
		self.construct_dataset(**kwargs)

	def fit(self, model = "PowerLaw", state_file="simple", save_state=True, **kwargs):
		"""
        Perform a simple fitting with a given model:
        PowerLaw, LogParabola, ...

        Args:
            model (str or gammapy.models): model name or function
                Default: "PowerLaw"
            state_file (str): state filename (pickle)
	        	Default: simple
	        save_state(bool)
	        	Default: True
	        **kwargs: passed to vtspy.model.default_model
        """

		if type(model) == str:
			spectral_model = default_model(model, **kwargs)
		elif hasattr(model, "tag"):
			spectral_model = model

		spectral_model = SkyModel(spectral_model=spectral_model, name=self.target_name)
		self.datasets.models = spectral_model
		self.stacked_dataset.models = [spectral_model]

		fit_joint = Fit()
		self.fit_results = fit_joint.run(datasets=[self.stacked_dataset])

		if self.fit_results.success:
			self._logging.info("Fit successfully.")
			if save_state:
				self.save_state(state_file)
				self._logging.info(f"The state is saved as '{state_file}'. You can load the state by vtspy.VeritasAnalysis('{state_file}').")
		else:
			self._logging.error("Fit failed.")

	def analysis(self, jobs=["sed"], state_file="analyzed", **kwargs):
		"""
		Perform a simple analysis, e.g., SED, lightcurve

		Args:
			jobs (list): list of jobs, 'sed', and/or 'lc'.
				Default: ['sed']
			state_file (str): state filename (pickle)
				Default: analyzed
			**kwargs: passed to vtspy.utils.define_time_intervals
		"""

		if type(jobs) == str:
			jobs = [jobs]

		if "sed" in jobs:
			self._logging.info("Generating flux points and SED...")
			self._energy_bins = kwargs.get("energy_bins", self._energy_bins)

			fpe = FluxPointsEstimator(
			    energy_edges=self.energy_bins,
			    source=self.target_name, selection_optional="all", **kwargs
			    )

			self.flux_points = fpe.run(datasets=[self.stacked_dataset])


			self._flux_points_dataset = FluxPointsDataset(
			    data=self.flux_points, models=self.stacked_dataset.models
			)
			self._logging.info("Completed.")

		if "lc" in jobs:
			self._logging.info("Generating lightcurve...")

			emin = kwargs.pop("emin", self.config["selection"]['emin'])
			emax = kwargs.pop("emax", self.config["selection"]['emax'])
			ul = kwargs.pop("ul", 2)
			tmin = kwargs.pop("tmin", self.config["selection"]['tmin'])
			tmax = kwargs.pop("tmax", self.config["selection"]['tmax'])

			time_intervals = utils.define_time_intervals(tmin, tmax, **kwargs)
			self._logging.info(f"The number of time intervals is {len(time_intervals)}")

			lc_maker = LightCurveEstimator(
			    energy_edges=[emin, emax] * u.TeV,
			    source=self.target_name,
			    time_intervals=time_intervals,
			    selection_optional="all"
			)

			self._lightcurve = lc_maker.run(self.datasets)
			self._lightcurve.sqrt_ts_threshold_ul = ul
			self._logging.info("Generating lightcurve is completed.")

		self.save_state(state_file)
		self._logging.info(f"The state is saved as '{state_file}'. You can load the state by vtspy.VeritasAnalysis('{state_file}').")

	def plot(self, output, **kwargs):
		"""
        Show various results: fit result, flux, SED, and lightcurve

        Args:
            output (str): a plot to show
                Options: ["roi", "fit", "flux", "sed", "lc"]
            filename (str): read the output (from FermiAnalysis.analysis)
        """
		if output == "roi":
			plot_ROI(veritas = self)
		elif output == "fit":
			self.stacked_dataset.plot_fit()
		elif output == "flux":
			ax = plt.gca()
			self.flux_points.plot(ax, sed_type="e2dnde", color="lightblue", label=self.target_name)
			self.flux_points.plot_ts_profiles(ax=ax, sed_type="e2dnde");
			ax.legend()
		elif output == "sed":
			kwargs_spectrum = {**kwargs, "kwargs_model": {"color":"blue", "label":"Pwl"}, "kwargs_fp":{"color":"blue", "marker":"o", "label":self.target_name}}
			kwargs_residuals = {"color": "blue", "markersize":4, "marker":'s'}
			ax_spec, ax_res = self._flux_points_dataset.plot_fit(kwargs_spectrum=kwargs_spectrum)
		elif output == "lc":
			self._lightcurve.plot(sed_type='eflux', label=self.target_name)

	def add_exclusion_region(self, coord = None, name=None, radius=0.3, update_dataset=False, **kwargs):
		"""
        Add exclusion region manually

        Args:
            coord (list, optioanl): [ra, dec]
            name (list, optional): [ra, dec]
            radius (float): size of an exlusion region
            update_dataset (bool): update dataset with the new exclusion region
        """
		if coord is not None:
			self._excluded_regions.append(CircleSkyRegion(
				center=SkyCoord(coord[0], coord[1], unit="deg", frame="icrs"),
				radius=radius * u.deg,))
		elif name is not None:
			src = SkyCoord.from_name(name)
			self._excluded_regions.append(CircleSkyRegion(
				center=SkyCoord(src.ra, src.dec, unit="deg", frame="icrs"),
				radius=radius * u.deg,))
		else:
			self._logging.error("Either coord or name should be provided.")
			return

		geom = WcsGeom.create(
		    	npix=(150, 150), binsz=0.05, skydir=self.target.galactic,
		    	proj="TAN", frame="icrs"
		)

		self._exclusion_mask = geom.region_mask(regions=self._excluded_regions, inside=False)

		if update_dataset:
			self._on_region = CircleSkyRegion(center=self.target, radius=Angle(np.sqrt(self._th2cut)*u.deg))
			self.construct_dataset(silent=True, **kwargs)

	def construct_dataset(self, **kwargs):
		"""
        Construct dataset for the gammapy analysis

        Args:
        	**kwargs: e.g., eff_cut, bias_cut, max_region_number, or others
        """

		self._eff_cut = kwargs.pop("eff_cut", self._eff_cut)
		self._bias_cut = kwargs.pop("bias_cut", self._bias_cut)
		self._max_region_number = kwargs.pop("max_region_number", self._max_region_number)

		datasets = Datasets()

		dataset_maker = SpectrumDatasetMaker(selection=["counts", "exposure", "edisp"], containment_correction=False)
		bkg_maker = ReflectedRegionsBackgroundMaker(exclusion_mask=self._exclusion_mask, max_region_number=self._max_region_number)

		if self._eff_cut > 0:
			safe_mask_maker_eff = SafeMaskMaker(methods=["aeff-max"], aeff_percent=self._eff_cut)
		else:
			safe_mask_maker_eff = None

		if self._bias_cut > 0:
			safe_mask_maker_bias = SafeMaskMaker(methods=["edisp-bias"], bias_percent=self._bias_cut)
		else:
			safe_mask_maker_bias = None

		geom = RegionGeom.create(region=self._on_region, axes=[self._energy_axis])
		dataset_empty = SpectrumDataset.create(
    		geom=geom, energy_axis_true=self._energy_axis_true
		)


		for obs_id, observation in zip(self._obs_ids, self.observations):
			dataset = dataset_maker.run(
				dataset_empty.copy(name=str(obs_id)), observation
			)

			dataset_on_off = bkg_maker.run(dataset, observation)

			if safe_mask_maker_eff is not None:
				dataset_on_off = safe_mask_maker_eff.run(dataset_on_off, observation)

			if safe_mask_maker_bias is not None:
				dataset_on_off = safe_mask_maker_bias.run(dataset_on_off, observation)

			datasets.append(dataset_on_off)

		self.datasets = datasets
		self.info_table = self.datasets.info_table()
		self.stacked_dataset = self.datasets.stack_reduce(name="veritas")
		self._N_off = sum(self.info_table["counts_off"])
		self._N_on = sum(self.info_table["counts"])
		self._alpha = np.average(self.info_table["alpha"], weights=1/self.info_table["livetime"])
		self._sigma = LiMaSiginficance(self._N_on, self._N_off, self._alpha)
		if not(kwargs.get("silent", False)):
			self._logging.info(r"N_on: {}, N_off: {}, alpha: {:.3f}, and sigma={:.1f}".format(self._N_on, self._N_off, self._alpha, self._sigma))

	def _exclusion_from_bright_src_list(self):
		srcfile = self.config["background"]["file"]
		distance = self.config["background"]["distance"]
		magnitude = self.config["background"]["magnitude"]
		ex_radius = self.config["selection"]["exc_radius"]

		bright_sources = utils.bright_source_list(srcfile)

		roi_cut = (abs(bright_sources[:,0]-self.target.ra.deg) < distance) \
        		* (abs(bright_sources[:,1]-self.target.dec.deg) < distance) \
        		* (bright_sources[:,2]+bright_sources[:,3] < magnitude)

		bright_sources = bright_sources[roi_cut]

		for src_pos in bright_sources:
			self._excluded_regions.append(CircleSkyRegion(
				center=SkyCoord(src_pos[0], src_pos[1], unit="deg", frame="icrs"),
				radius=ex_radius * u.deg,))

		geom = WcsGeom.create(
			npix=(150, 150), binsz=0.05, skydir=self.target.galactic,
			proj="TAN", frame="icrs"
		)

		return geom.region_mask(regions=self._excluded_regions, inside=False)

	def _exclusion_from_simbad(self):

		## under testing

		try:
			from astroquery.simbad import Simbad
		except:
			self._logging.error("A simbad module cannot be found. pip install astroquery.")
			return

		distance = self.config["background"]["distance"]
		ex_radius = self.config["selection"]["exc_radius"]
		magnitude = self.config["background"]["magnitude"]

		simbad = Simbad()
		simbad.reset_votable_fields()
		simbad.add_votable_fields('ra', 'dec', "flux(B)", "flux(V)", "jp11")
		simbad.remove_votable_fields('coordinates')

		self._logging.info("Querying bright sources within FoV with Simbad.")
		srcs_tab = simbad.query_region(self._on_region.center, radius=distance*u.deg)
		srcs_tab = srcs_tab[srcs_tab["FLUX_B"]<magnitude]
		srcs_tab = srcs_tab[srcs_tab["FLUX_V"]!=np.ma.masked]
		self._logging.info(f"{len(srcs_tab)} sources have been found.")

		srcs = SkyCoord(['{} {}'.format(src["RA"], src["DEC"]) for src in srcs_tab], unit=(u.hourangle, u.deg))
		self._excluded_regions = [CircleSkyRegion(center=src, radius=ex_radius * u.deg,) for src in srcs]

		geom = WcsGeom.create(
		    npix=(150, 150), binsz=0.05, skydir=self.target.galactic,
		    proj="TAN", frame="icrs"
		)

		return geom.region_mask(regions=self._excluded_regions, inside=False)

	def _quick_check_runlist(self):
		run_list = fits.open(self._datadir+"/obs-index.fits.gz")

		run_list = run_list[1].data["OBS_ID"]

		run_exist = []
		for file in glob.glob(self._datadir+"/*.anasum.fits"):
		    run_exist.append(re.findall("([0-9]+)", file)[0])

		return (len(run_exist) == len(run_list))



from setuptools import setup, find_packages
from pathlib import Path


## read __version__
with open(Path(__file__).parent.absolute().joinpath('vtspy/version.py')) as f:
    exec(f.read())


setup(name='vtspy',
      version=__version__,
      description="VERITAS analysis tool",
      packages=find_packages(),
      install_requires=['astropy',
                        'numpy',
                        'matplotlib>=3.5.2',
                        'gammapy',
                        'uproot',
                        'html2text',
                        'astroquery',
                        ],
      extras_require={'tests': ['pytest', 'pytest-ordering'],
                      'examples': ['ipywidgets', 'ipympl', 'nodejs']
                      },
      package_data={
        'vtspy': ['refdata/*.fit', 'refdata/*.dat']
      },
      include_package_data=True,
      author='Donggeun Tak',
      author_email='donggeun.tak@gmail.com',
      url='https://vtspy.readthedocs.io/en/latest/',
      long_description="",
      use_scm_version={
          "write_to": Path(__file__).parent.joinpath("vtspy/version.py"),
          "write_to_template": "__version__ = '{version}'",
      },
      )



import os
import glob

import yaml

import uproot

from astropy.io import fits

from . import utils
from .utils import logger, SCRIPT_DIR

from pathlib import Path

class JointConfig:
	"""
	This is to generate the configuration file compatible to
	the Fermipy configuration. The input file is the VERITAS
	event file from EventDisplay. The format can be either
	`.root' or `.fits'.

	Args:
	    files (str): a input file or directory containing
	    	root or fits files (root or fits)
	    config_file (str): Fermi config filename (yaml)
	    	Default: config.yaml
	    verbosity (int)
	    **kwargs: passed to JointConfig.init
	"""

	def __init__(self, files=None, config_file="config.yaml", verbosity=1, **kwargs):
		self._logging = logger(verbosity=verbosity)

		self._filename = config_file
		path = Path(config_file)
		if path.is_file() and (files is None):
			config = self.get_config(config_file)
			self.fermi_config = config.pop("fermi")
			self.veritas_config = config.pop("veritas")
			self.print_info()
			self._logging.info(f'a configuration file ({config_file}) is loaded.')
		else:
			self.init(files=files, config_file=config_file, **kwargs)
			self.print_info(config_file=config_file)
			self._logging.info(f'a configuration file ({config_file}) is created.')

	def init(self, files=None, config_file="config.yaml", **kwargs):
		"""
		Generate a default configuration file
		Args:
			files (str): a input file or directory containing
		    	root or fits files (root or fits)
		    config_file (str): Fermi config filename (yaml)
		    	Default: config.yaml
			fermi_outdir (str): path to the Fermi-LAT output directory
		        Default: ./fermi/
		    fermi_data (str): path to the Fermi-LAT data directory
	        	Default: ./fermi/
	        veritas_outdir (str): path to the VERITAS output directory
	        	Default: ./veritas/
			veritas_data (str): path to the VERITAS data directory
	        	Default: ./veritas/
	        joint_outdir (str): path to the joint-fit output directory
	        	Default: ./joint/
	        **kwargs
        """

		fermi_outdir = kwargs.pop("fermi_outdir", "./fermi/")
		fermi_data = kwargs.pop("fermi_data", "./fermi/")
		veritas_outdir = kwargs.pop("veritas_outdir", "./veritas/")
		veritas_data = kwargs.pop("veritas_data", "./veritas/")
		joint_outdir = kwargs.pop("joint_outdir", "./joint/")

		gald = kwargs.pop("gald", "gll_iem_v07.fits")
		iso = kwargs.pop("iso", "iso_P8R3_SOURCE_V3_v1.txt")
		info = kwargs.pop("info", {})

		if files is None:
			filelist = glob.glob(veritas_data+"/*")
		else:
			filelist = glob.glob(files+"/*")

		for file in filelist:
			if ".gz" in file:
				filelist.remove(file)

		self.fermi_config = self._empty4fermi(outdir = fermi_outdir, datadir = fermi_data, gald=gald, iso=iso)
		self.veritas_config = self._empty4veritas(outdir = veritas_outdir, datadir = veritas_data)

		self.veritas_config["data"]["anasum"] = os.path.dirname(filelist[0])

		info = {**info,
			'fermi':{
			'selection':{
				'ra': None,
				'dec': None,
				'tmin': None,
				'tmax':None}},
			'veritas':{
			'selection':{
				'ra': None,
				'dec': None,
				'tmin': None,
				'tmax': None,
			}
			}
		}

		ra = info['fermi']['selection']['ra']
		dec = info['fermi']['selection']['dec']

		for file in filelist:

			if file!=None:

				if '.root' in file:

					anasum = uproot.open(file)

					tRun = anasum['total_1']['stereo']['tRunSummary']

					if info['fermi']['selection']['ra'] == None:
						ra = float(tRun['TargetRAJ2000'].arrays(library="np")['TargetRAJ2000'][0])
					else:
						temp = float(tRun['TargetRAJ2000'].arrays(library="np")['TargetRAJ2000'][0])
						if temp != ra:
							self._logging.error("[Error] RA values in input files are different.")

					if dec == None:
						dec = float(tRun['TargetDecJ2000'].arrays(library="np")['TargetDecJ2000'][0])
					else:
						temp = float(tRun['TargetDecJ2000'].arrays(library="np")['TargetDecJ2000'][0])
						if temp != dec:
							self._logging.error("[Error] DEC values in input files are different.")

					tmin_mjd = tRun['MJDOn'].arrays(library="np")['MJDOn'][0]
					tmin_utc = utils.MJD2UTC(tmin_mjd)
					tmin = utils.UTC2MET(tmin_utc[:10])

					tmax_mjd = tRun['MJDOff'].arrays(library="np")['MJDOff'][0]
					tmax_utc = utils.MJD2UTC(tmax_mjd)
					tmax = utils.UTC2MET(tmax_utc[:10])+60*60*24

					target = tRun["TargetName"].arrays(library="np")['TargetName'][0]

				elif 'anasum.fits' in file:

					header = fits.open(file)[1].header
					if ra == None:
						ra = header['RA_OBJ']
					else:
						temp = header['RA_OBJ']
						if temp != ra:
							self._logging.error("[Error] RA values in input files are different.")

					if dec == None:
						dec = header['DEC_OBJ']
					else:
						temp = header['DEC_OBJ']
						if temp != dec:
							self._logging.error("[Error] DEC values in input files are different.")

					tmin_utc = header['DATE-OBS']
					tmin = utils.UTC2MET(tmin_utc[:10])
					tmin_mjd = utils.UTC2MJD(tmin_utc)

					tmax_utc = header['DATE-END']
					tmax = utils.UTC2MET(tmax_utc[:10])+60*60*24
					tmax_mjd = utils.UTC2MJD(tmax_utc)

					target = header['OBJECT']
				else:
					continue

				if info['fermi']['selection']['tmin'] is not None:
					if info['fermi']['selection']['tmin'] < tmin:
						tmin = info['fermi']['selection']['tmin']
					else:
						info['fermi']['selection']['tmin'] = tmin

				if info['fermi']['selection']['tmax'] is not None:
					if info['fermi']['selection']['tmax'] > tmax:
						tmax = info['fermi']['selection']['tmax']
					else:
						info['fermi']['selection']['tmax'] = tmax

				if info['veritas']['selection']['tmin'] is not None:
					if info['veritas']['selection']['tmin'] < tmin_mjd:
						tmin_mjd = info['veritas']['selection']['tmin']
					else:
						info['veritas']['selection']['tmin'] = tmin_mjd
				if info['veritas']['selection']['tmax'] is not None:
					if info['veritas']['selection']['tmax'] > tmax_mjd:
						tmax_mjd = info['veritas']['selection']['tmax']
					else:
						info['veritas']['selection']['tmax'] = tmax_mjd


				info = {**info,
					'fermi':{
					'selection':{
						'ra': ra,
						'dec': dec,
						'tmin': tmin,
						'tmax': tmax,}
						#'target': target},
						},
					'veritas':{
					'selection':{
						'ra': ra,
						'dec': dec,
						'tmin': tmin_mjd,
						'tmax': tmax_mjd,
						'target': target,
					}}
					}



		info['fermi'] = self._filter(self.fermi_config, info['fermi'])
		info['veritas'] = self._filter(self.veritas_config, info['veritas'])

		self.fermi_config = self._update(self.fermi_config, info['fermi'])
		self.veritas_config = self._update(self.veritas_config, info['veritas'])
		self.joint_config = {'fileio':{
								'outdir': joint_outdir,
							}}
		info = {"fermi": self.fermi_config, "veritas": self.veritas_config, "joint": self.joint_config}

		self.set_config(info, config_file)

		self.config = info

	def change_time_interval(self, tmin, tmax, scale = "utc", instrument="all"):
		"""
		Change and update a time interval

		Args:
		tmin (float or str): start time
		tmax (float or str): end time
		scale (str): "utc", "mjd", or "met"
			Default: "utc"
		instrument (str): "fermi", "veritas", or "all"
			Default: "all"


		"""
		if scale.lower() == "utc":
			tmin_mjd = utils.UTC2MJD(tmin)
			tmax_mjd = utils.UTC2MJD(tmax)
			tmin_met= utils.UTC2MET(tmin[:10])
			tmax_met = utils.UTC2MET(tmax[:10])+60*60*24
		elif scale.lower() == "mjd":
			tmin = float(tmin)
			tmax = float(tmax)
			tmin_mjd = tmin
			tmax_mjd = tmax
			tmin_utc = utils.MJD2UTC(tmin)
			tmax_utc = utils.MJD2UTC(tmax)
			tmin_met = utils.UTC2MET(tmin_utc[:10])
			tmax_met = utils.UTC2MET(tmax_utc[:10])+60*60*24
		elif scale.lower() == "met":
			tmin = float(tmin)
			tmax = float(tmax)
			tmin_met = tmin
			tmax_met = tmax
			tmin_utc = utils.MET2UTC(tmin)
			tmax_utc = utils.MET2UTC(tmax)
			tmin_mjd = utils.UTC2MJD(tmin_utc)
			tmax_mjd = utils.UTC2MJD(tmax_utc)
		else:
			self._logging.error("The input 'scale' parameter is not 'MJD', 'MET', or 'UTC'.")
			return

		if instrument.lower() == "fermi" or instrument.lower() == "all":
			self.fermi_config["selection"]["tmin"] = tmin_met
			self.fermi_config["selection"]["tmax"] = tmax_met

		if instrument.lower() == "veritas" or instrument.lower() == "all":
			self.veritas_config["selection"]["tmin"] = tmin_mjd
			self.veritas_config["selection"]["tmax"] = tmax_mjd

		self.set_config(self.config, self._filename)
		self.print_info(self._filename)


	@staticmethod
	def get_config(config_file="config.yaml"):
		"""
	    Read a config file.

	    Args:
	        config_file (str): Fermi config filename (yaml)
	        	Default: config.yaml
	    """
		return yaml.load(open(config_file), Loader=yaml.FullLoader)

	@classmethod
	def print_config(self, config_file="config.yaml"):
		"""
	    print a config file.

	    Args:
	    	config_file (str): Fermi config filename (yaml)
				Default: config.yaml
	    """
		self.config = self.get_config(config_file)
		self.fermi_config = self.config.get("fermi")
		self.veritas_config = self.config.get("veritas")

		if not(hasattr(self, "_logging")):
			self._logging = logger()
		self._logging.info("\n"+yaml.dump(self.config, sort_keys=False, default_flow_style=False))

	@classmethod
	def print_info(self, config_file="config.yaml"):
		self.config = self.get_config(config_file)
		self.fermi_config = self.config.pop("fermi")
		self.veritas_config = self.config.pop("veritas")
		self._logging = logger()
		self._logging.info("-"*20+" Info "+"-"*20)
		self._logging.info("target: {}".format(self.veritas_config["selection"]["target"]))
		self._logging.info("localization:")
		self._logging.info("\t(ra, dec) : ({}, {})".format(self.veritas_config["selection"]["ra"],
		                                    self.veritas_config["selection"]["dec"]))
		self._logging.info("\t(glat, glon) : ({}, {})".format(self.veritas_config["selection"]["glat"],
		                                    self.veritas_config["selection"]["glon"]))
		self._logging.info("time interval:")
		self._logging.info("\tveritas : {} - {}".format(utils.MJD2UTC(self.veritas_config["selection"]["tmin"]),
		                                    utils.MJD2UTC(self.veritas_config["selection"]["tmax"])))
		self._logging.info("\tfermi : {} - {}".format(utils.MET2UTC(self.fermi_config["selection"]["tmin"]),
		                                  utils.MET2UTC(self.fermi_config["selection"]["tmax"])))
		self._logging.info("-"*45)

	@staticmethod
	def set_config(info, config_file="config.yaml"):
		"""
	    Write inputs into a config file.

	    Args:
	    	info (dict): overwrite the input info into a config file
	        config_file (str): Fermi config filename (yaml)
	        	Default: config.yaml
	    """
		with open(config_file, "w") as f:
			yaml.dump(info, f)


	@classmethod
	def update_config(self, info, instrument, config_file="config.yaml"):
		"""
	    Update a config file.

	    Args:
	    	info (dict): update info in a config file
	    	instrument (str): either fermi or veritas
	        config_file (str): Fermi config filename (yaml)
				Default: config.yaml
	    """
		pre_info = self.get_config(config_file)

		info = self._filter(pre_info[instrument], info)

		pre_info[instrument] = self._update(pre_info[instrument], info)

		self.set_config(pre_info, config_file)

	@staticmethod
	def _filter(pre_info, info):
		if len(info) != 0:
			for key in list(info.keys()):
				for subkey in list(info[key].keys()):
					if (pre_info[key][subkey] == info[key][subkey]) or (info[key][subkey]==None):
						info[key].pop(subkey)
		return info

	@staticmethod
	def _update(pre_info, info):
		if len(info) != 0:
			for key in info.keys():
				for subkey in info[key].keys():
					pre_info[key][subkey] = info[key][subkey]
					if (key == "selection") and (subkey=="ra"):
						if info['selection']['ra'] != None and info['selection']['dec'] != None:
							if 'binning' in pre_info.keys():
								pre_info['binning']['coordsys'] = 'CEL'
							glon, glat = utils.CEL2GAL(info['selection']['ra'], info['selection']['dec'])
							pre_info['selection']['glon'], pre_info['selection']['glat'] = float(glon), float(glat)

					if (key == "selection") and (subkey=="glon"):
						if info['selection']['glon'] != None and info['selection']['glat'] != None:
							if 'binning' in pre_info.keys():
								pre_info['binning']['coordsys'] = 'CEL'
							ra, dec = utils.GAL2CEL(info['selection']['glon'], info['selection']['glat'])
							pre_info['selection']['ra'], pre_info['selection']['dec'] = float(ra), float(dec)

		return pre_info

	@staticmethod
	def _empty4fermi(outdir = "./fermi/", datadir = "./fermi/", gald = "gll_iem_v07.fits", iso = "iso_P8R3_SOURCE_V3_v1.txt"):
		if not(os.path.isdir(outdir)):
			os.system(f"mkdir {outdir}")
		if not(os.path.isdir(datadir)):
			os.system(f"mkdir {datadir}")
		if not(os.path.isdir(f"{outdir}/log")):
			os.system(f"mkdir {outdir}/log")
			os.system(f": > {outdir}/log/fermipy.log")

		info = {
 				'data': {
 					'evfile': f"{datadir}/EV00.lst",
 					'scfile': f"{datadir}/SC00.fits",
 					'ltcube': None
 					},
 				'binning': {
 					'roiwidth': 12,
  					'binsz': 0.08,
  					'binsperdec': 8,
  					'coordsys': None,
  					'projtype': 'WCS',
  					},
 				'selection': {
 					'emin': 100,
					'emax': 300000,
					'tmin': None,
					'tmax': None,
					'zmax': 90,
					'evclass': 128,
					'evtype': 3,
					'glon': None,
					'glat': None,
					'ra': None,
					'dec': None,
					'target': None
					},
				'gtlike': {
					'edisp': True,
					'irfs': 'P8R3_SOURCE_V3',
					'edisp_disable': ['isodiff', 'galdiff']
					},
				'model': {
					'src_roiwidth': 15,
					'galdiff': f'$FERMI_DIFFUSE_DIR/{gald}',
					'isodiff': f'$FERMI_DIFFUSE_DIR/{iso}',
					'catalogs': SCRIPT_DIR+'/refdata/gll_psc_v22.fit'
					},
				'fileio': {
					'outdir' : outdir,
   					'logfile' : f"{outdir}/log/fermipy.log",
					'usescratch': False
					},
				}
		return info

	@staticmethod
	def _empty4veritas(outdir = "./veritas/", datadir="./veritas/"):
		if not(os.path.isdir(outdir)):
			os.system(f"mkdir {outdir}")

		info = {
			'background':
			{
				'file': SCRIPT_DIR+"/refdata/Hipparcos_MAG8_1997.dat",
				'distance': 1.75,
				'magnitude': 7,
				'simbad': True,
			},
			'data': {
				'anasum': datadir,
			},
			'fileio':{
				'outdir': outdir,
			},
			'cuts':{
				'th2cut': 0.008,
				'eff_cut': 0,
				'bias_cut': 0,
			},
			'selection':
			{
				'target': None,
				'ra': None,
				'dec': None,
				'tmin' : None,
				'tmax' : None,
				'emin': 0.1,
				'emax': 10,
				'nbin': 6,
				'format': "mjd",
				'max_region_number': 6,
				'radius': 2.0,
				'exc_on_region_radius': 0.7,
				'exc_radius': 0.25,
				},

			}
		return info
